from __future__ import annotations
import enum as _enum
import typing as _typing
import datetime as _datetime
import re as _re
from google.cloud import bigquery as _bigquery
from google.cloud import storage as _storage
from google.oauth2 import service_account as _service_account
from toolbox.bigquery_sink.utils import nest as _nest


class Options(object):
    def __init__(
        self,
        project_id: str,
        dataset_id: str = None,
        temp_bucket_name: str = None,
        temp_bucket_root_path: str = None,
        service_account_credentials: dict = None,
        bq_location: str = None,
        correlation_id: str = None,
        now: _datetime.datetime = None,
        labels: dict = None,
    ):
        """
        Reduce copy paste code: create an access config once and reuse it for multiple sinks
        :param project_id: The project id where table data will be stored
        :param dataset_id: The dataset where table data will be stored
        :param temp_bucket_name: The google cloud storage bucket where temp files are "parked"
        :param temp_bucket_root_path: The path to the root folder where temp files will be "parked" inside the temp bucket
        :param service_account_credentials: You can provide service account credentials for non-user bound access to google cloud
        :param bq_location: The location where the data should be hosted (only relevant if table does not exist, yet).
        :param correlation_id: (optional) identifier that allows finding associated objects in microservice architecture
        :param now: (optional) allows trace passed time since a process was originally started
        :param labels: A dictionary of labels that should be attached to tables and other resources
        """
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.temp_bucket_name = temp_bucket_name
        self.temp_bucket_root_path = temp_bucket_root_path
        self.service_account_credentials = service_account_credentials
        self.bq_location = bq_location
        self.labels = labels

        if correlation_id:
            if not _re.match(r'[a-z]+', correlation_id):
                raise ValueError('Correlation id must be sequence of lower case characters!')

        self.correlation_id = correlation_id
        self.now = now

    def get_bigquery_client(self):
        return self._get_client(cls=_bigquery.Client,)

    def get_storage_client(self):
        return self._get_client(cls=_storage.Client,)

    def _get_client(self, cls):
        if self.service_account_credentials:
            project_id = self.service_account_credentials["project_id"]
        else:
            project_id = self.project_id

        credentials = _service_account.Credentials.from_service_account_info(
            self.service_account_credentials
        )
        return cls(project=project_id, credentials=credentials)

    def replace(
        self,
        project_id: str = None,
        dataset_id: str = None,
        temp_bucket_name: str = None,
        temp_bucket_root_path: str = None,
        service_account_credentials: dict = None,
        bq_location: str = None,
        correlation_id: str = None,
        now: _datetime.datetime = None,
        labels: dict = None,
    ):
        """
        Return an access config with the same values but replacing the values that are not None
        :param project_id: The project id where table data will be stored
        :param dataset_id: The dataset where table data will be stored
        :param temp_bucket_name: The google cloud storage bucket where temp files are "parked"
        :param temp_bucket_root_path: The path to the root folder where temp files will be "parked" inside the temp bucket
        :param service_account_credentials: You can provide service account credentials for non-user bound access to google cloud
        :param bq_location: The location where the data should be hosted (only relevant if table does not exist, yet).
        :param correlation_id: (optional) identifier that allows finding associated objects in microservice architecture
        :param now: (optional) allows trace passed time since a process was originally started
        :param labels: A dictionary of labels that should be attached to tables and other resources
        :return: A copy of the access config after replacing the provided fields
        """
        return Options(
            project_id=project_id or self.project_id,
            dataset_id=dataset_id or self.dataset_id,
            temp_bucket_name=temp_bucket_name or self.temp_bucket_name,
            temp_bucket_root_path=temp_bucket_root_path or self.temp_bucket_root_path,
            service_account_credentials=service_account_credentials
            or self.service_account_credentials,
            bq_location=bq_location or self.bq_location,
            correlation_id=correlation_id or self.correlation_id,
            now=now or self.now,
            labels=labels,
        )


class FieldType(_enum.Enum):
    STRING = "STRING"
    BYTES = "BYTES"

    BOOLEAN = "BOOLEAN"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    NUMERIC = "NUMERIC"

    TIMESTAMP = "TIMESTAMP"
    DATE = "DATE"
    DATETIME = "DATETIME"  # note that this ignores timezone! Use rather timestamp and set to UTC
    TIME = "TIME"

    STRUCT = "STRUCT"
    RECORD = "STRUCT"


class FieldMode(_enum.Enum):
    NULLABLE = "NULLABLE"
    REQUIRED = "REQUIRED"
    REPEATED = "REPEATED"  # define an array type in standard sql


class SourcePathElements(_enum.Enum):
    """
    Use these to define special fields in the source path of a schema field
    """

    ROOT = object()  # make the path absolute and not relative
    LIST_INDEX = object()  # mark that you're accessing a list within the source path


class _MissingToken(object):
    """
    Used internally to detect missing values
    """

    pass


source_fn_type = _typing.Callable[[_typing.Any, _typing.List], _typing.Any]


class SchemaField(object):
    """
    Used to build schemas for bigquery tables
    """

    def __init__(
        self,
        name: str,
        field_type: FieldType,
        description: str = None,
        source_path: _typing.List[str] = None,
        source_fn: source_fn_type = None,
        fields: _typing.List[SchemaField] = None,
        mode: FieldMode = None,
    ):
        """
        Create new schema field
        :param name: Name of the field in bigquery
        :param field_type: Type of field, pick from any bigquery compatible type, e.g. STRING, TIMESTAMP, INTEGER, FLOAT
        :param description: Describe the content of this field
        :param source_path: Can be used to easily extract data from a data source row. Should be a list of keys
        :param fields: If the field_type is RECORD you can specify fields for the "sub document" of the record
        :param mode: pick from NULLABLE, REQUIRED, REPEATED (view bigquery docs for meaning)
        """
        self.name = name
        self.field_type = field_type
        self.description = description
        self.source_path = source_path
        self.source_fn = source_fn
        self.fields = fields
        self.mode = mode or FieldMode.NULLABLE

    def to_bq_field(self):
        """
        Creates a SchemaField compatible with bigquery's python library
        :return: SchemaField from original google lib
        """
        return _bigquery.schema.SchemaField(
            name=self.name,
            field_type=self.field_type.value,
            mode=self.mode.value or "NULLABLE",
            description=self.description,
            fields=[f.to_bq_field() for f in self.fields or ()],
        )

    def extract(self, row, should_ensure_type=True, should_fire_exception=None):
        """
        Extracts the field value from row given source_path
        :param row: the row to extract from
        :param should_ensure_type: Should the types be casted into the output types?
        :param should_fire_exception:
            Pass in fn to check whether an exception should be fired. fn(row, path, exception) -> Boolean
        :return: The field value that was extracted
        """

        return self._extract_inner(
            row=row,
            path=[],
            should_ensure_type=should_ensure_type,
            should_fire_exception=should_fire_exception,
        )

    def _extract_inner(self, row, path, should_ensure_type, should_fire_exception):
        """
        Helper function of extract.
        Extracts the value from the row according to the source definition on the schema field and the provided path.
        :param row: The original row to extract data from
        :param path:
            This function is used recursively and path provides information where we are reading from in the source
        :param should_ensure_type: Should the types be casted into the output types?
        :param should_fire_exception:
            Pass in fn to check whether an exception should be fired. fn(row, path, exception) -> Boolean
        :return: Returns the value of this field extracted from the row
        """
        try:
            missing_token = _MissingToken()
            if self.source_fn:
                return self.source_fn(row, path)

            source_path = self._create_source_path(path=path)

            if self.mode == FieldMode.REPEATED:
                return self._extract_inner_repeated(
                    row, source_path, should_ensure_type, should_fire_exception
                )

            if self.field_type == FieldType.STRUCT:
                value = {
                    field.name: field._extract_inner(
                        row,
                        path=source_path,
                        should_ensure_type=should_ensure_type,
                        should_fire_exception=should_fire_exception,
                    )
                    for field in self.fields
                }
            else:
                value = _nest.get(obj=row, path=source_path, default=missing_token)
                if value == missing_token:
                    value = None
                if should_ensure_type:
                    value = self._ensure_type(value=value)

            return value
        except Exception as exception:
            if should_fire_exception:
                should_fire = should_fire_exception(row, path, exception)
                if should_fire:
                    raise
            else:
                raise

    def _create_source_path(self, path):
        if self.source_path is not None:
            if len(self.source_path) and self.source_path[0] == SourcePathElements.ROOT:
                source_path = self.source_path[1:]

                # replace LIST_INDEX elements in source path, if they follow the same path!
                for idx, (source_path_item, path_item) in enumerate(
                    list(zip(source_path, path))
                ):
                    if source_path_item == path:
                        continue
                    if source_path_item == SourcePathElements.LIST_INDEX:
                        source_path[idx] = path_item
                    if source_path_item != path_item:
                        break
            else:
                source_path = path + self.source_path
        else:
            source_path = path + [self.name]

        return source_path

    def _extract_inner_repeated(
        self, row, source_path, should_ensure_type, should_fire_exception
    ):
        if SourcePathElements.LIST_INDEX in source_path:
            # if there is a LIST_INDEX value in the path,
            # split up the path and deal with the remainder recursively
            list_index_position = source_path.index(SourcePathElements.LIST_INDEX)
            path_to = source_path[:list_index_position]
            remaining_path = source_path[list_index_position + 1 :]
            inner_list = _nest.get(row, path_to)
            next_inner_list = []

            if SourcePathElements.LIST_INDEX in remaining_path:
                # if in the remainder is another LIST_INDEX, we expect to get a list back
                for idx in range(len(inner_list)):
                    next_inner_list += self.replace(
                        source_path=remaining_path
                    )._extract_inner(
                        row,
                        path=path_to + [idx],
                        should_ensure_type=should_ensure_type,
                        should_fire_exception=should_fire_exception,
                    )
                return next_inner_list
            else:
                # if there is no other LIST_INDEX remaining, we expect to get back individual values
                # therefore we remove the field mode REPEATED in the recursive call
                for idx in range(len(inner_list)):
                    next_inner_list.append(
                        self.replace(
                            mode=FieldMode.NULLABLE, source_path=remaining_path
                        )._extract_inner(
                            row,
                            path=path_to + [idx],
                            should_ensure_type=should_ensure_type,
                            should_fire_exception=should_fire_exception,
                        )
                    )
                return next_inner_list

        inner_list = _nest.get(row, source_path)
        if not inner_list or not isinstance(inner_list, list):
            return []

        if self.field_type == FieldType.STRUCT:
            value = [
                {
                    field.name: field._extract_inner(
                        row,
                        path=source_path + [idx],
                        should_ensure_type=should_ensure_type,
                        should_fire_exception=should_fire_exception,
                    )
                    for field in self.fields
                }
                for idx in range(len(inner_list))
            ]
        else:
            value = []
            for idx, inner_value in enumerate(inner_list):
                if should_ensure_type:
                    value.append(self._ensure_type(value=inner_value))
                else:
                    value.append(inner_value)
        return value

    def _ensure_type(self, value):
        """
        Helper function of self.extract - ensures that the output types are correct.
        The types are prepared so that BigQuery can import them!
        :param value: Input value that should be checked and casted
        :return: A type-casted value that BigQuery should be able to read it
        """
        if isinstance(value, float) and self.field_type == FieldType.NUMERIC:
            value = str(round(value, 8))

        if isinstance(value, int) and self.field_type == FieldType.BOOLEAN:
            value = value != 0

        if isinstance(value, _datetime.datetime) and self.field_type == FieldType.DATE:
            value = value.date()

        if (
            not isinstance(value, str)
            and value is not None
            and self.field_type == FieldType.STRING
        ):
            value = str(value)

        if self.field_type == FieldType.INTEGER and value is not None:
            value = int(value)

        if self.field_type == FieldType.FLOAT and value is not None:
            value = float(value)

        if self.field_type == FieldType.DATE and isinstance(value, int):
            value = _datetime.datetime.fromtimestamp(value).date()

        if (
            self.field_type == FieldType.DATETIME
            or self.field_type == FieldType.TIMESTAMP
        ) and isinstance(value, int):
            value = _datetime.datetime.utcfromtimestamp(value)

        return value

    def replace(
        self,
        name: str = None,
        field_type: FieldType = None,
        description: str = None,
        source_path: _typing.List[str] = None,
        source_fn: source_fn_type = None,
        fields: _typing.List[SchemaField] = None,
        mode: FieldMode = None,
    ):
        """
        Returns a copy of this schema field but replacing the values with provided parameters (if not None)
        :param name: Name of the field in bigquery
        :param field_type: Type of field, pick from any bigquery compatible type, e.g. STRING, TIMESTAMP, INTEGER, FLOAT
        :param description: Describe the content of this field
        :param source_path: Can be used to easily extract data from a data source row. Should be a list of keys
        :param fields: If the field_type is RECORD you can specify fields for the "sub document" of the record
        :param mode: pick from NULLABLE, REQUIRED, REPEATED (view bigquery docs for meaning)
        """
        return SchemaField(
            name=name if name is not None else self.name,
            field_type=field_type if field_type is not None else self.field_type,
            description=description if description is not None else self.description,
            source_path=source_path if source_path is not None else self.source_path,
            source_fn=source_fn if source_fn is not None else self.source_fn,
            fields=fields if fields is not None else self.fields,
            mode=mode if mode is not None else self.mode,
        )


def create_view(name: str, query: str, options: Options):
    """
    Create a view inside bigquery

    :param name: The name of the view (corresponds to a table id)
    :param query: The SQL query that specifies the content of the vieww
    :param options: The access config object that defines project_id and other common parameters
    """
    bigquery = options.get_bigquery_client()
    view_ref = "{}.{}.{}".format(options.project_id, options.dataset_id, name)

    view = _bigquery.Table(view_ref)
    view.view_query = query
    return bigquery.create_table(view, exists_ok=True)


if __name__ == "__main__":
    pass



import re as _re
import json as _json
import datetime as _datetime
import decimal as _decimal
import tempfile as _tempfile
import contextlib as _contextlib
import enum as _enum
import typing as _typing

from google.cloud import bigquery as _bigquery

from toolbox.bigquery_sink.utils import generate_id as _generate_id
from toolbox import bigquery_sink as _bigquery_sink


def create_table_date_partitioning(field, expiration_ms=None):
    """
    Utility function to create date(time) partitioned tables
    """
    return {
        "type": "time_partitioning",
        "definition": _bigquery.TimePartitioning(
            type_=_bigquery.TimePartitioningType.DAY,
            field=field,
            expiration_ms=expiration_ms,
        ),
    }


class WriteDisposition(_enum.Enum):
    """
    Allows controlling whether table should be appended/overwritten/only written to if empty

    APPEND: append to the table (creates table if it does not exist)
    IF_EMPTY: will only write if the table is empty (or does not exist)
    REPLACE: replaces table content in atomic operation
    """

    APPEND = _bigquery.WriteDisposition.WRITE_APPEND
    IF_EMPTY = _bigquery.WriteDisposition.WRITE_EMPTY
    REPLACE = _bigquery.WriteDisposition.WRITE_TRUNCATE


class BQBulkSink(object):
    """
    Data Sink for loading chunks of rows into bigquery (not streaming insert!)
    """

    def __init__(
        self,
        table_id: str,
        options: _bigquery_sink.Options,
        table_partitioning: dict = None,
        table_partition_date: _datetime.date = None,
        table_description: str = None,
        schema: _typing.List[_bigquery_sink.SchemaField] = None,
        write_disposition: WriteDisposition = WriteDisposition.APPEND,
        auto_update_table_schema: bool = False,
    ):
        """
        :param table_id: the table id where the data should be stored. This should not contain project_id or dataset_id
        :param dataset_id: the dataset id where the destination table is located, you can overwrite the dataset_id provided from the options
        :param project_id: the project id where the destination table is located, you can overwrite the project_id provided from the options
        :param table_partitioning: BigQuery table partitioning information, use `create_table_date_partitioning` method to create correct values
        :param table_partition_date: If you want to (over-) write a specific partition, you can specify a date(time) object. This only works for date partitioned tables. If you pass in a datetime object, it will still replace the entire DATE partition.
        :param table_description: You can provide a description of the table.
        :param schema: The schema of the table
        :param write_disposition: Specify whether you want to only write if the table is empty or append or replace table content
        :param auto_update_table_schema: Should the schema be automatically updated when uploading content?
        """

        self.options = options
        self.project_id = options.project_id
        self.dataset_id = options.dataset_id
        self.table_id = table_id
        self.table_ref = "{}.{}.{}".format(
            self.project_id, self.dataset_id, self.table_id
        )
        self.table_partitioning = table_partitioning
        self.table_partition_date = table_partition_date
        self.table_description = table_description
        self.schema = schema
        self.bq_schema = (
            None if self.schema is None else [f.to_bq_field() for f in self.schema]
        )
        self.labels = options.labels or {}
        self.write_disposition = write_disposition.value
        self.auto_update_table_schema = auto_update_table_schema
        self.bq_location = options.bq_location or "US"
        self.bigquery = options.get_bigquery_client()
        self.storage = options.get_storage_client()
        self.temp_bucket_name = options.temp_bucket_name
        self.temp_bucket_root_path = options.temp_bucket_root_path
        self.now = options.now or _datetime.datetime.utcnow()
        self.correlation_id = options.correlation_id
        self.rows_written = 0

    @_contextlib.contextmanager
    def open(self):
        """
        1) creates temp file
        2) allows writing data using context manager into newline delimited json format
        3) after moving out of the context will start submitting:
        3.1) uploads file to google storage
        3.2) loads bq table from google storage (and ensures that dataset & table exist in bq)
        :return: None
        """
        with _tempfile.TemporaryFile() as tmp_file:

            def __write(row):
                to_write = _json.dumps(row, default=self._json_default_fn) + "\n"
                tmp_file.write(to_write.encode("utf-8"))

            yield __write

            self._create_bq_dataset(exists_ok=True)  # ensures that dataset exists
            table = self._create_bq_table(exists_ok=True)  # ensures that table exists

            storage_uri = self._upload_file_obj_to_storage(file_obj=tmp_file)
            self._load_bq_table_from_storage(storage_uri=storage_uri, table=table)

    def from_iterable(
        self,
        iterable,
        force_values=None,
        should_ensure_type=True,
        should_fire_exception=False,
    ):
        """
        Read from an iterable and directly upload.
        Allows providing force_values to add static values in addition to the rows coming from the source.
        Fields from the source are fed through the `field.extract` method to convert them according to the schema.
        :param iterable: The data source which is read row by row
        :param force_values: Provide a dict of key value pairs that is going to be written into the sink for each row
        :param should_ensure_type: whether the types should be cast so that BigQuery can understand them
        :param should_fire_exception: whether exceptions should be fired or caught silently
        :return: Nr of rows written
        """
        rows_written = 0
        with self.open() as sink_write:
            for row in iterable:
                to_write = {}
                for field in self.schema:
                    to_write[field.name] = field.extract(
                        row,
                        should_ensure_type=should_ensure_type,
                        should_fire_exception=should_fire_exception,
                    )

                if force_values:
                    for key, val in force_values.items():
                        to_write[key] = val
                sink_write(to_write)
                rows_written += 1

        self.rows_written += rows_written
        return rows_written

    def from_query(self, query):
        self._create_bq_dataset(exists_ok=True)  # ensures that dataset exists
        self._create_bq_table(exists_ok=True)  # ensures that table exists
        if self.table_partition_date:
            table_ref = "{}${:%Y%m%d}".format(self.table_ref, self.table_partition_date)
        else:
            table_ref = self.table_ref

        job_config = _bigquery.QueryJobConfig(
            allow_large_results=True,
            destination=table_ref,
            write_disposition=self.write_disposition,
        )
        query_job = self.bigquery.query(
            query=query, job_config=job_config, job_id=self._generate_job_id(),
        )
        result = query_job.result()
        self.rows_written = result.total_rows

    def create_related_view(self, name, sql_template):
        if "{TABLE}" in sql_template:
            sql_template = sql_template.format(TABLE="`{}`".format(self.table_ref))

        return _bigquery_sink.create_view(
            name=name,
            sql=sql_template,
            dataset_id=self.dataset_id,
            options=self.options,
        )

    def _generate_job_id(self):
        job_id = "{runner}--{project}--{dataset}--{table}--{date}-{time}-{correlation_id}-{random}".format(
            runner="lh-dwh-etl",
            project=self.project_id,
            dataset=self.dataset_id,
            table=self.table_id,
            date=self.now.strftime("%Y-%m-%d"),
            time=self.now.strftime("%H-%M-%S"),
            correlation_id=self.correlation_id,
            random=_generate_id.generate_id(),
        )
        return job_id

    def _json_default_fn(self, obj):
        """
        Ensure that we can json dump also special types: datetime, date and decimal
        Override this fn to work with even more types
        """
        if isinstance(obj, (_datetime.datetime, _datetime.date)):
            return obj.isoformat()

        if isinstance(obj, _decimal.Decimal):
            return str(obj)

    def _upload_file_obj_to_storage(self, file_obj, rewind=True):
        """
        Upload content of a file_obj into google cloud storage.

        :param file_obj: The file object to upload
        :param rewind: Whether or not to rewind the provided file
        :return: the google cloud storage uri (e.g. 'gs://BUCKET/FILE_PATH')
        """
        bucket = self.storage.get_bucket(bucket_or_name=self.temp_bucket_name)

        if self.temp_bucket_root_path:
            root_path = self.temp_bucket_root_path
            # remove leading or trailing '/' chars
            root_path = _re.match(r"^/?(.*)/?$", root_path).group(1)
            if root_path:
                root_path += "/"
        else:
            root_path = ""

        file_path = "{root_path}{project}/{dataset}/{table}/{date}/{time}-{correlation}-{random}.nljson".format(
            root_path=root_path,
            project=self.project_id,
            dataset=self.dataset_id,
            table=self.table_id,
            date=self.now.strftime("%Y-%m-%d"),
            time=self.now.strftime("%H-%M-%S"),
            correlation=self.correlation_id,
            random=_generate_id.generate_id(),
        )

        blob = bucket.blob(file_path)
        blob.upload_from_file(file_obj=file_obj, rewind=rewind)

        return "gs://{}/{}".format(self.temp_bucket_name, file_path)

    def _create_bq_dataset(self, exists_ok):
        """
        Creates a dataset in bigquery for the given project_id and dataset_id.
        Dataset location is automatically EU
        :param exists_ok: Whether or not this should complain if dataset already exists
        :return: the google lib dataset object
        """
        dataset = _bigquery.Dataset(
            dataset_ref="{}.{}".format(self.project_id, self.dataset_id)
        )
        dataset.location = self.bq_location
        dataset = self.bigquery.create_dataset(dataset=dataset, exists_ok=exists_ok)
        return dataset

    def _create_bq_table(self, exists_ok):
        """
        1) Creates a table in bigquery given the project_id, dataset_id & table_id.
        2) Compares new vs old schema and updates if required (can be controlled with `self.auto_update_table_schema`)
        :param exists_ok: Whether or not this should complain if table already exists
        :return: the google lib table object
        """
        table = _bigquery.Table(table_ref=self.table_ref, schema=self.bq_schema)
        if self.table_partitioning:
            setattr(
                table,
                self.table_partitioning["type"],
                self.table_partitioning["definition"],
            )
        table = self.bigquery.create_table(table=table, exists_ok=exists_ok)

        self._check_and_update_labels(table=table)
        self._check_and_update_description(table=table)
        self._check_and_update_schema(table=table)

        if self.table_partition_date:
            table = _bigquery.Table(
                table_ref="{}${:%Y%m%d}".format(
                    self.table_ref, self.table_partition_date
                ),
                schema=self.bq_schema,
            )

        return table

    def _check_and_update_labels(self, table):
        """
        Make sure the labels are up2date on the table
        """
        if table.labels != self.labels:
            labels = {
                k: None  # make sure "old" labels are actually removed
                for k in table.labels.keys()
            }
            labels.update(self.labels)
            table.labels = labels
            self.bigquery.update_table(table=table, fields=["labels"])

    def _check_and_update_description(self, table):
        """
        Make sure the description is up2date
        """
        if table.description != self.table_description:
            table.description = self.table_description
            self.bigquery.update_table(table=table, fields=["description"])

    def _check_and_update_schema(self, table):
        """
        Make sure that the schema of table matches the one provided.
        The ordering is NOT taken into account - changes in order will not be updated
        """

        def extract_schema_fields(schema):
            if not schema:
                return None
            return sorted(
                [
                    (
                        f.name.upper(),
                        f.field_type.upper(),
                        f.mode.upper(),
                        f.description,
                        extract_schema_fields(f.fields),
                    )
                    for f in schema
                ]
            )

        current_schema_repr = extract_schema_fields(table.schema)
        new_schema_repr = extract_schema_fields(self.bq_schema)
        if current_schema_repr != new_schema_repr:
            if self.auto_update_table_schema:
                table.schema = self.bq_schema
                self.bigquery.update_table(table=table, fields=["schema"])
            else:
                raise ValueError(
                    "Schema changed {}.{}.{}".format(
                        self.project_id, self.dataset_id, self.table_id
                    )
                )

    def _load_bq_table_from_storage(self, storage_uri, table):
        """
        Create/Update a bigquery table given a google cloud storage uri
        :param storage_uri: The source uri where to get the data from, e.g. 'gs://BUCKET/FILE_PATH'
        :return: None
        """
        job_config = _bigquery.LoadJobConfig(
            schema=self.bq_schema,
            source_format=_bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
            write_disposition=self.write_disposition,
        )
        load_job = self.bigquery.load_table_from_uri(
            source_uris=storage_uri,
            destination=table,
            job_config=job_config,
            job_id=self._generate_job_id(),
        )
        load_job.result()  # Waits for table load to complete.


if __name__ == "__main__":
    pass



"""
Code adapted from by: https://github.com/dgilland/pydash
"""

import collections


def get(obj, path, default=None):
    current = obj
    for key in path:
        if isinstance(current, list) and isinstance(key, int):
            if key < 0 or key >= len(current):
                return default
            else:
                current = current[key]
                continue
        if current is None or key not in current:
            return default
        current = current[key]

    return current


def in_obj(obj, path):
    if len(path) == 0:
        return False

    if len(path) == 1:
        current = obj
    else:
        current = get(obj, path[:-1])

    key = path[-1]
    if isinstance(obj, list) and isinstance(key, int):
        if 0 <= key < len(obj):
            return True
        else:
            return False

    return key in current


def insert(container, key_path, item):
    """
    >>> insert({}, ['a', '1', '2', 'world'], 'hello')
    {'a': {'1': {'2': {'world': 'hello'}}}}
    """
    if isinstance(container, collections.OrderedDict):
        gen = collections.OrderedDict
        update = lambda i, k, v: i.update({k: v})
    else:
        gen = dict
        update = lambda i, k, v: i.__setitem__(k, v)

    sub_container = container
    for key in key_path[:-1]:
        if isinstance(key, int):
            raise ValueError('No int keys allowed in deep insert')
        if key not in sub_container:
            update(sub_container, key, gen())
        sub_container = sub_container[key]

    update(sub_container, key_path[-1], item)

    return container



if __name__ == '__main__':
    pass